#custom-parameter {
    position: absolute;
    z-index: 999999;
    top: 0;
    left: 0;
    display: flex;
    flex-direction: column;
    align-items: center;
    color: white;
    width: 100%;
}

#custom-parameter>form {
    background-color: rgba(15, 1, 28, 0.8);
    padding: 20px;
    display: none;
    width: 100%;
    border: 1px solid rgba(255, 255, 255, 0.65);
}

#custom-parameter.toggled>form {
    display: flex;
    flex-wrap: wrap;
    justify-content: center;
}

#custom-parameter>form>.input {
    margin: 5px;
    padding: 5px;
    display: flex;
    flex-direction: column;
    align-items: center;
    justify-content: center;
    background-color: rgba(255, 255, 255, 0.2);
}

#custom-parameter>form>.input>label {
    font-size: 18px;
    margin-bottom: 7px;
}

#custom-parameter>button {
    all: unset;
    cursor: pointer;
    background-color: transparent;
    border: 1px solid white;
    border-bottom-left-radius: 5px;
    border-bottom-right-radius: 5px;
    border-top: none;
    opacity: 0.65;
    display: flex;
    width: 40px;
    height: 20px;
    align-items: center;
    justify-content: center;
}

#custom-parameter.toggled>button {
    background-color: rgb(15, 1, 28);
    opacity: 0.8;
}

#custom-parameter>button:hover {
    background-color: rgba(255, 255, 255, 0.4);
}

#custom-parameter>button>img {
    transform: rotateZ(180deg);
}

#custom-parameter.toggled>button>img {
    transform: rotateZ(0deg);
}

#custom-parameter>form>.submit {
    width: 100%;
    display: flex;
    justify-content: center;
    margin-top: 10px;
}

#custom-parameter>form>.submit>button[type="submit"] {
    all: unset;
    cursor: pointer;
    background-color: transparent;
    padding: 7px 14px;
    border-radius: 3px;
    border: 1px solid rgba(255, 255, 255, 0.5);
    background-color: rgba(255, 255, 255, 0.1);
}

#custom-parameter>form>.submit>button[type="submit"]:hover {
    background-color: rgba(255, 255, 255, 0.3);
}

/* loading ui */

.loader {
    -webkit-animation: loader-turn 1s linear infinite;
    animation: loader-turn 1s linear infinite;
    padding: 1rem;
    max-width: 60px;
    width: 100%;
}

@-webkit-keyframes loader-turn {
    50% {
        transform: rotate(180deg);
    }

    100% {
        transform: rotate(720deg);
    }
}

@keyframes loader-turn {
    50% {
        transform: rotate(180deg);
    }

    100% {
        transform: rotate(720deg);
    }
}

.loader__value {
    -webkit-animation: loader-stroke 6s linear infinite;
    animation: loader-stroke 6s linear infinite;
    fill: none;
    stroke-dasharray: 63;
    stroke-dashoffset: 63;
    stroke-linecap: round;
    stroke-width: 3;
}

.loader__value:nth-child(1) {
    stroke: white;
}

.loader__value:nth-child(2) {
    stroke: white;
    -webkit-animation-delay: 1s;
    animation-delay: 1s;
}

.loader__value:nth-child(3) {
    stroke: white;
    -webkit-animation-delay: 2s;
    animation-delay: 2s;
}

.loader__value:nth-child(4) {
    stroke: white;
    -webkit-animation-delay: 3s;
    animation-delay: 3s;
}

.loader__value:nth-child(5) {
    stroke: white;
    -webkit-animation-delay: 4s;
    animation-delay: 4s;
}

.loader__value:nth-child(6) {
    stroke: white;
    -webkit-animation-delay: 5s;
    animation-delay: 5s;
}

@-webkit-keyframes loader-stroke {
    8.3333333333% {
        stroke-dashoffset: 0;
    }

    16.6666666667%,
    100% {
        stroke-dashoffset: 63;
    }
}

@keyframes loader-stroke {
    8.3333333333% {
        stroke-dashoffset: 0;
    }

    16.6666666667%,
    100% {
        stroke-dashoffset: 63;
    }
}